""" Be prepared. """

import math

import pygame.display
import pygame.surface
import pygame.locals as pgl

import pymunk
import pymunk.pygame_util

from .base import BaseScene
from ..battery import BatteryManager
from ..lights import LightManager, light_fitting_by_type, check_space_for_light
from ..infobar import InfoBar
from ..obstacles import ObstacleManager
from ..events import SceneChangeEvent
from ..utils import debug_timer, shadowed_text, write_save_file
from ..loader import loader
from ..transforms import Overlay, Alpha, ColourWedges

from ..constants import SCREEN_SIZE, FONTS, FPS, NIGHT_HOURS_PER_TICK, DEBUG
from ..widgets import ImageButton
from ..turnip import Turnip, TurnipInvalidPosition


class DayScene(BaseScene):

    BRIGHTNESS = Overlay(colour=(255, 255, 255, 50))

    def enter(self, gamestate):
        self._space = pymunk.Space()
        self._obstacles = ObstacleManager(self._space, gamestate)
        self._lights = LightManager(self._space, gamestate)
        self._battery = BatteryManager(gamestate)
        self._infobar = InfoBar("day", battery=self._battery, scene=self)
        self._turnips = []
        self._paused = False
        self._tool = None
        self._light_colors = None
        self._dragging = None
        # Turnip
        self.grow_turnips(gamestate)
        # Tools
        self._light_toolbar = []
        self._tools = self.create_tools(gamestate)
        # Background
        self._soil = loader.load_image(
            "textures", "soil.png", transform=self.BRIGHTNESS)
        # Check if we've lost
        self._game_over_text = []
        if gamestate.seeds == 0 and len(self._turnips) == 0:
            self._draw_you_lose(gamestate)
        elif gamestate.harvested >= gamestate.turnip_target:
            self._draw_you_win(gamestate)
        else:
            write_save_file(gamestate.serialize())
        self._ending = False

    def _draw_you_lose(self, gamestate):
        overlay = pygame.surface.Surface(
            (SCREEN_SIZE[0], 240), pgl.SWSURFACE).convert_alpha()
        overlay.fill((0, 0, 0, 128))
        self._game_over_text.append((overlay, (0, 250)))
        self._game_over_text.append(
            (shadowed_text("You Lost", FONTS["bold"], 48), (400, 280)))
        self._game_over_text.append(
            (shadowed_text("You have no seeds and no turnips growing",
                           FONTS["sans"], 24), (300, 350)))
        self._game_over_text.append(
            (shadowed_text("Press a key to return to the menu",
                           FONTS["sans"], 24), (350, 400)))

    def _draw_you_win(self, gamestate):
        overlay = pygame.surface.Surface(
            (SCREEN_SIZE[0], 240), pgl.SWSURFACE).convert_alpha()
        overlay.fill((0, 0, 0, 128))
        self._game_over_text.append((overlay, (0, 250)))
        self._game_over_text.append(
            (shadowed_text("You Win", FONTS["bold"], 48), (400, 280)))
        self._game_over_text.append(
            (shadowed_text(
                ("You have Successfully Harvested %d turnips" %
                 gamestate.harvested),
                FONTS["sans"], 24),
             (300, 350)))
        self._game_over_text.append(
            (shadowed_text("Press a key to return to the menu",
                           FONTS["sans"], 24), (350, 400)))

    def grow_turnips(self, gamestate):
        for turnip_data in gamestate.turnips:
            turnip = Turnip(space=self._space, **turnip_data)
            # Turnips grow at dawn
            seeds = turnip.grow()
            if seeds:
                gamestate.seeds += seeds
                gamestate.harvested += 1
            else:
                self._turnips.append(turnip)

    def create_tools(self, gamestate):
        tools = []

        x, y, step = 50, SCREEN_SIZE[1] - 40, 50

        tools.append(ImageButton('32', 'seed.png', name='seed', pos=(x, y)))
        x += step

        for light_config in gamestate.station["available_lights"]:
            tool = ImageButton(
                '32', '%s.png' % light_config["type"], name='light',
                pos=(x, y))
            tool.light_config = light_config
            tools.append(tool)
            x += step

        tools.append(ImageButton(
            '32', 'remove.png', name='remove light', pos=(x, y)))
        x += step

        tools.append(ImageButton(
            '32', 'default_cursor.png', name='reset tool', pos=(x, y)))

        tools.append(ImageButton(
            '32', 'night.png', name='start night',
            pos=(SCREEN_SIZE[0] - 100, y)))
        tools.append(ImageButton(
            '32', 'exit.png', name='exit', pos=(SCREEN_SIZE[0] - 50, y)))
        return tools

    def exit(self, gamestate):
        self._unset_cursor()
        turnip_data = [turnip.serialize() for turnip in self._turnips]
        gamestate.turnips = turnip_data

    def end_day(self, gamestate):
        if self._ending:
            return
        self._battery.apply_recharge()
        gamestate.update_lights(self._lights)
        self._ending = True
        from .night import NightScene
        SceneChangeEvent.post(scene=NightScene())

    @property
    def turnip_count(self):
        return len(self._turnips)

    @property
    def power_usage(self):
        power = self._lights.total_power_usage()
        power = power / (FPS * NIGHT_HOURS_PER_TICK)
        return int(round(power))

    @debug_timer("day.render")
    def render(self, surface, gamestate):
        surface.blit(self._soil, (0, 0))

        for turnip in self._turnips:
            turnip.render(surface)
        self._lights.render_light(surface)
        self._obstacles.render(surface)
        self._lights.render_fittings(surface)
        self._infobar.render(surface, gamestate)
        for tool in self._tools:
            tool.render(surface)
        for light_tool in self._light_toolbar:
            light_tool.render(surface)
        self._draw_cursor(surface)
        if self._game_over_text:
            for surf, pos in self._game_over_text:
                surface.blit(surf, pos)

    def _draw_light_toolbar(self, light_config, x):
        height = SCREEN_SIZE[1] - 80
        self._light_toolbar = []
        colour_combos = light_config["available_colours"]
        for combo in colour_combos:
            colours = combo.split("/")
            light_fitting = light_fitting_by_type(light_config["type"])
            light_tool = ImageButton(
                "32", light_fitting, transform=ColourWedges(colours=colours),
                pos=(x, height), name=combo)
            light_tool.colours = colours
            self._light_toolbar.append(light_tool)
            x += 40

    def _clear_light_toolbar(self):
        self._light_toolbar = []

    def _place_seed(self, gamestate, ev):
        if gamestate.seeds > 0:
            # plant seed
            # We don't want top-left to equal the mouse position,
            # since that looks weird, but we don't want to center
            # the turnip under the mouse either, since that
            # causes issues as well, so we compromise
            pos = (ev.pos[0] - 18, ev.pos[1] - 18)
            try:
                turnip = Turnip(age=0, pos=pos, space=self._space)
                self._turnips.append(turnip)
                gamestate.seeds -= 1
            except TurnipInvalidPosition:
                # TODO: Add error sound or something
                pass

    def _update_light_angle(self, pos, gamestate):
        # Update the angle of the given light
        pos = pymunk.pygame_util.to_pygame(pos, pygame.display.get_surface())
        distance = pos - self._dragging.position
        angle = math.atan2(distance[1], distance[0])
        # Set light angle to this position
        self._dragging.ray_manager.direction = math.degrees(angle)
        # Hackily update gamestate with new angle
        for light_cfg in gamestate.station["lights"]:
            light_pos = pymunk.Vec2d(light_cfg["position"])
            if light_pos.get_dist_sqrd(self._dragging.position) < 5.0:
                light_cfg["direction"] = math.degrees(angle)
                break

    def _place_light(self, gamestate, cfg, colours, ev):
        cfg = cfg.copy()
        cost = cfg.pop("cost")
        cfg.pop("available_colours")
        if gamestate.seeds > cost:
            pos = pymunk.pygame_util.from_pygame(
                ev.pos, pygame.display.get_surface())
            # Bail if we're too close to an existing light, obstacle or turnip
            if check_space_for_light(self._space, pos, max_distance=25):
                return
            gamestate.seeds -= cost
            cfg["position"] = pos
            cfg["colours"] = colours
            gamestate.station["lights"].append(cfg)
            self._lights.add_light(cfg)

    def _remove_light(self, ev):
        light = self._lights.nearest(ev.pos, surfpos=True, max_distance=25.0)
        if light:
            self._lights.remove_light(light)

    def event(self, ev, gamestate):
        if self._ending:
            return
        if self._game_over_text:
            if ev.type in (pgl.KEYDOWN, pgl.MOUSEBUTTONDOWN):
                self._ending = True
                from .menu import MenuScene
                SceneChangeEvent.post(scene=MenuScene())
        if ev.type == pgl.KEYDOWN:
            if ev.key in (pgl.K_q, pgl.K_ESCAPE):
                self._ending = True
                from .menu import MenuScene
                SceneChangeEvent.post(scene=MenuScene())
            elif ev.key == pgl.K_e:
                self.end_day(gamestate)
            elif ev.key == pgl.K_SPACE and DEBUG:
                self._paused = not self._paused
        elif ev.type == pgl.MOUSEBUTTONDOWN:
            if ev.button == 1:
                # Check tools
                for tool in self._tools:
                    if tool.pressed(ev):
                        self._color = None
                        if tool.name == 'reset tool':
                            self._unset_cursor()
                            self._tool = None
                            self._clear_light_toolbar()
                        elif tool.name == 'start night':
                            self.end_day(gamestate)
                        elif tool.name == 'exit':
                            self._ending = True
                            from .menu import MenuScene
                            SceneChangeEvent.post(scene=MenuScene())
                        else:
                            self._tool = tool
                            if self._tool.name == 'seed':
                                self._set_cursor(
                                    'seed', transform=Alpha(alpha=172))
                                self._clear_light_toolbar()
                            elif self._tool.name == 'remove light':
                                self._set_cursor(
                                    'remove', transform=Alpha(alpha=172))
                                self._clear_light_toolbar()
                            elif self._tool.name == 'light':
                                self._unset_cursor()
                                self._draw_light_toolbar(
                                    self._tool.light_config, 100)
                        return
                # Check light toolbar
                for light_tool in self._light_toolbar:
                    if light_tool.pressed(ev):
                        fitting_image = light_fitting_by_type(
                            self._tool.light_config["type"])
                        self._set_cursor(
                            fitting_image[:-4],  # strip .png
                            transform=ColourWedges(colours=light_tool.colours))
                        # colour=COLOURS[0] + (172,)))
                        self._light_colors = light_tool.colours
                        return
                if self._tool:
                    if self._tool.name == "seed":
                        self._place_seed(gamestate, ev)
                    elif self._tool.name == "remove light":
                        self._remove_light(ev)
                    elif self._tool.name == "light" and self._light_colors:
                        self._place_light(
                            gamestate, self._tool.light_config,
                            self._light_colors, ev)
                else:
                    # Not tool, so check lights
                    self._lights.toggle_nearest(ev.pos, surfpos=True)
            elif ev.button == 3:
                light = self._lights.nearest(ev.pos, surfpos=True,
                                             max_distance=20.0)
                if light:
                    # Start drag to rotate light
                    self._dragging = light
                elif self._tool:
                    # Unset tool
                    self._tool = None
                    self._clear_light_toolbar()
                    self._unset_cursor()
        elif ev.type == pgl.MOUSEMOTION:
            if self._dragging:
                # Calculate angle between current position and mouse pos
                self._update_light_angle(ev.pos, gamestate)
        elif ev.type == pgl.MOUSEBUTTONUP:
            self._dragging = None

    @debug_timer("day.tick")
    def tick(self, gamestate):
        if not self._paused:
            self._lights.tick()
